;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "javelin" "0.2.3"
  "Implementation of harpoon: bookmarks on steroids."
  '((emacs "28.1"))
  :url "https://github.com/DamianB-BitFlipper/javelin.el"
  :commit "e58733d6dfb3282d5b6ad4d295e8c71a8e5abc71"
  :revdesc "e58733d6dfb3"
  :keywords '("tools" "languages"))
